/*
 * sun-rbootd   yamamori@kt.rim.or.jp
 */

#include <unistd.h>
#include <fcntl.h>
#include <stdio.h>

#include <sys/param.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <net/if.h>
#include <netinet/in.h>
#include <netinet/if_ether.h>

#include <netdb.h>


extern char *ether_interfaces(void);
extern int ether_open(char *name);
extern int ether_write(int fd, char *pkt, int len);
extern int ether_read(int fd, char *pkt, int len);


typedef unsigned char u_int8_t;
typedef unsigned short u_int16_t;

#include "rmp.h"
#undef RMP_MAX_PACKET
#define RMP_MAX_PACKET 1514
#include "rmp_var.h"

#define TFTPBOOT "/tftpboot"

#define u_word_to_ul(v) (((v).val[0]<<16)|(v).val[1])


static struct rmp_packet pb;


static int
my_ether_write(int fd, char *pkt, int len)
{
  if (len > 1500) {
    return 1;
  }
  pb.hp_hdr.len = len;

  ether_write(fd, pkt, len);
  return 0;
}


int
main()
{
  int fd_e, fd_boot;
  int size;
  int pid;
  unsigned long offset;
  struct hostent *hp;
  char boot_file[8+1];
  char hostname[MAXHOSTNAMELEN];

  if (chdir(TFTPBOOT) < 0) {
    perror(TFTPBOOT);
    return 1;
  }

  if ((fd_e = ether_open(ether_interfaces())) < 0) {
    perror("ether_open");
    return 1;
  }

  if ((pid = fork()) < 0) {
    perror("fork");
    return 1;
  }
  if (pid) {
    return 0;
  }
  setsid();


  for (;;) {
    if (ether_read(fd_e, (char *)&pb, 60) <= 0) {
      continue;
    }

    *(struct ether_addr *)pb.hp_hdr.daddr =
      *(struct ether_addr *)pb.hp_hdr.saddr;

    pb.hp_llc.dxsap = HPEXT_SXSAP;
    pb.hp_llc.sxsap = HPEXT_DXSAP;

    switch (pb.r_type) {
      case RMP_BOOT_REQ:
	if (
	  pb.r_brq.rmp_session == RMP_PROBESID &&
	  u_word_to_ul(pb.r_brq.rmp_seqno) == 0
	) {
	  pb.r_brpl.rmp_flnmsize = 6;
	  strcpy(&pb.r_brpl.rmp_flnm, "server");
	} else {
	  pb.r_brpl.rmp_flnmsize = 4;
	  strcpy(&pb.r_brpl.rmp_flnm, "BOOT");
	}

	pb.r_brpl.rmp_type = RMP_BOOT_REPL;
	pb.r_brpl.rmp_retcode = 0;
	pb.r_brpl.rmp_session = 0;
	pb.r_brpl.rmp_version = RMP_VERSION;

	my_ether_write(fd_e, (char *)&pb, 46);
	break;

      case RMP_READ_REQ:
	size = pb.r_rrq.rmp_size;
	offset = u_word_to_ul(pb.r_rrq.rmp_offset);

	pb.r_rrpl.rmp_type = RMP_READ_REPL;
	pb.r_rrpl.rmp_retcode = 0;

	if (
	  ether_ntohost(hostname, (struct ether_addr* )pb.hp_hdr.saddr) != 0
	) {
	  fprintf(
	    stderr, "%s unknown\n",
	    ether_ntoa((struct ether_addr*)pb.hp_hdr.saddr)
	  );
	  break;
	}
	if ((hp = gethostbyname(hostname)) == NULL) {
	  fprintf(stderr, "%s: unknown\n", hostname);
	  break;
	}
	sprintf(boot_file, "%08X", *(long *)hp->h_addr);

	if ((fd_boot = open(boot_file, O_RDONLY, 0)) < 0) {
	  perror(boot_file);
	  break;
	}
        lseek(fd_boot, offset, SEEK_SET);
	read(fd_boot, &pb.r_rrpl.rmp_data, size);
	close(fd_boot);

	my_ether_write(fd_e, (char *)&pb, size + 18);
	break;
    }
  }
}
